﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the top two items are equal, or FALSE otherwise
  /// </summary>
  /// <typeparam name="T">Stacktype</typeparam>
  public abstract class EqualsExpression<T> : StatelessExpression {
    public bool Eval(IStack<T> stack, IStack<bool> booleanStack) {
      if (stack.Count < 2) return false;

      var items = stack.Pop(2);

      booleanStack.Push(items[0].Equals(items[1]));

      return true;
    }
  }

  [PushExpression(StackTypes.Integer, "INTEGER.=", StackTypes.Boolean)]
  public class IntegerEqualsExpression : EqualsExpression<long> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.IntegerStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Float, "FLOAT.=", StackTypes.Boolean)]
  public class FloatEqualsExpression : EqualsExpression<double> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Boolean, "BOOLEAN.=", StackTypes.Boolean)]
  public class BooleanEqualsExpression : EqualsExpression<bool> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.BooleanStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Name, "NAME.=", StackTypes.Boolean)]
  public class NameEqualsExpression : EqualsExpression<string> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.NameStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.=", StackTypes.Boolean)]
  public class ExecEqualsExpression : EqualsExpression<Expression> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.ExecStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Code, "CODE.=", StackTypes.Boolean)]
  public class CodeEqualsExpression : EqualsExpression<Expression> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.CodeStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.=", StackTypes.Boolean)]
  public class CharEqualsExpression : EqualsExpression<char> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.CharStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.String, "STRING.=", StackTypes.Boolean)]
  public class StringEqualsExpression : EqualsExpression<string> {
    public override bool Eval(IPushInterpreter interpreter) {
      return Eval(interpreter.StringStack, interpreter.BooleanStack);
    }
  }
}